"""
The package of the partially-observed time-series imputation method linear interpolation.
"""

# Created by Cole Sussmeier <colesussmeier@gmail.com>
# License: BSD-3-Clause

from .model import Lerp

__all__ = [
    "Lerp",
]
