"""

"""

# Created by Wenjie Du <wenjay.du@gmail.com>
# License: BSD-3-Clause

from ai4ts.client import TimeSeriesAI

from ..utils.logging import logger


class Gungnir(TimeSeriesAI):
    def __init__(self):
        super().__init__()
        logger.info(
            "The functionalities of client Gungnir have not been implemented yet,\n"
            "which will be updated later 2024 once our close beta test is ready.\n"
            "AI for real-world time series is coming. 🚀 Stay tuned please! https://time-series.ai"
        )
